/*
 * Structure definition.
 *
 * Copyright (C) 2019, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "config.h"
#include "declaration.h"
#include "structure.h"
#include "templates.h"

/* Generate an enumeration of operator codes. */

void write_opcode_definition(struct interface *iface, FILE *fp)
{
  fprintf(fp, opcode_enumeration_prologue, iface->name);
  write_opcodes(iface->signatures, fp, iface->name);
  fputs(opcode_enumeration_epilogue, fp);
}

/* Generate operator code enumeration entries. */

void write_opcodes(struct signature *sig, FILE *fp, const char *name)
{
  char *opcode;

  if (sig == NULL)
    return;

  fprintf(fp, opcode_identifier, name, sig->operation);

  /* Obtain any explicit opcode from the signature attributes. */

  opcode = get_opcode(sig->attributes);
  if (opcode != NULL)
    fprintf(fp, " = %s", opcode);

  fputs(",\n", fp);

  write_opcodes(sig->tail, fp, name);
}

/* Generate structures for messages corresponding to the signatures. */

void write_structures(struct signature *sig, FILE *fp, struct interface *iface)
{
  if (sig == NULL)
    return;

  write_structure(sig, fp, iface);
  write_structures(sig->tail, fp, iface);
}

/* Terminate a structure definition, setting any explicit structure
   alignment. */

static void write_structure_alignment(FILE *fp)
{
  char alignment[32];

  if (conf.alignment)
    sprintf(alignment, structure_epilogue_alignment, conf.alignment);
  else
    strcpy(alignment, "");

  fprintf(fp, structure_epilogue, alignment);
}

/* Generate message structures for a signature. */

void write_structure(struct signature *sig, FILE *fp, struct interface *iface)
{
  struct parameter *param = sig->parameters;
  char *opname = get_operation_name(iface, sig),
       *protocol = get_protocol(iface->attributes),
       *opcode_type;
  int input_words, input_items, output_words, output_items;

  count_parameters(param, &input_words, &input_items, &output_words, &output_items);

  /* Adjust the input words where a protocol is used as the opcode. */

  if (protocol != NULL)
    input_words++;

  /* Generate the structures for populating and interpreting messages. */

  if (input_words)
  {
    fprintf(fp, structure_prologue, "in", opname);
    if (protocol != NULL)
    {
      opcode_type = get_opcode_type(sig->attributes);

      if (opcode_type != NULL)
        fprintf(fp, structure_opcode_member, opcode_type);
      else
        fprintf(fp, structure_opcode_member, L4_OPCODE_TYPE);
    }
    write_declarations(param, IN_PARAMETER, WORD_CLASS, fp);
    write_structure_alignment(fp);
  }

  if (output_words)
  {
    fprintf(fp, structure_prologue, "out", opname);
    write_declarations(param, OUT_PARAMETER, WORD_CLASS, fp);
    write_structure_alignment(fp);
  }

  /* Free allocated strings. */

  free(opname);
}
